using System.Collections.ObjectModel;
using System.Drawing;
using System.Text.Json;
using View_by_Distance.Shared.Models;
using View_by_Distance.Shared.Models.Properties;
using View_by_Distance.Shared.Models.Stateless;

namespace View_by_Distance.Distance.Models.Stateless;

internal static class FilterLogicB
{

    private static void MoveUnableToMatch(FilePath filePath)
    {
        string checkFile = $"{filePath.FullName}.unk";
        if (File.Exists(filePath.FullName) && !File.Exists(checkFile))
            File.Move(filePath.FullName, checkFile);
    }

    private static void LocationContainersParallelFor(DistanceSettings distanceSettings, ICompareSettings compareSettings, List<LocationContainer> locationContainers, ExifDirectory exifDirectory)
    {
        string? json;
        if (exifDirectory.FilePath.Id is null)
            return;
        DateOnly dateOnly = DateOnly.FromDateTime(new DateTime(exifDirectory.FilePath.CreationTicks));
        int? wholePercentages = IMapping.GetWholePercentages(compareSettings, exifDirectory.FilePath);
        if (wholePercentages is null)
        {
            if (distanceSettings.DistanceMoveUnableToMatch)
                MoveUnableToMatch(exifDirectory.FilePath);
            return;
        }
        json = Metadata.Models.Stateless.IMetadata.GetOutputResolution(exifDirectory);
        if (json is null || !json.Contains(nameof(DateTime)))
        {
            if (distanceSettings.DistanceMoveUnableToMatch)
                MoveUnableToMatch(exifDirectory.FilePath);
            return;
        }
        FaceFile? faceFile = JsonSerializer.Deserialize(json, FaceFileGenerationContext.Default.FaceFile);
        if (faceFile is null || faceFile.Location is null)
        {
            if (distanceSettings.DistanceMoveUnableToMatch)
                MoveUnableToMatch(exifDirectory.FilePath);
            return;
        }
        RectangleF? rectangle = Shared.Models.Stateless.ILocation.GetPercentagesRectangle(distanceSettings, wholePercentages.Value);
        if (rectangle is null)
            return;
        LocationContainer locationContainer = new(CreationDateOnly: dateOnly,
                                                  ExifDirectory: exifDirectory,
                                                  Encoding: exifDirectory.Encoding,
                                                  FaceFile: faceFile,
                                                  FilePath: exifDirectory.FilePath,
                                                  LengthPermyriad: null,
                                                  LengthSource: null,
                                                  PersonKeyFormattedAndKeyTicksAndDisplayDirectoryName: exifDirectory.PersonKeyFormattedAndKeyTicksAndDisplayDirectoryName,
                                                  Rectangle: rectangle,
                                                  WholePercentages: wholePercentages);
        lock (locationContainers)
            locationContainers.Add(locationContainer);
    }

    internal static ReadOnlyCollection<LocationContainer> GetLocationContainers(DistanceSettings distanceSettings, ICompareSettings compareSettings, ICompare compare, ReadOnlyCollection<ExifDirectory> exifDirectories, string sourceClass)
    {
        List<LocationContainer> results = [];
        int maxDegreeOfParallelism = compareSettings.MaxDegreeOfParallelism;
        int totalSeconds = (int)Math.Floor(new TimeSpan(DateTime.Now.Ticks - compare.Ticks).TotalSeconds);
        ParallelOptions parallelOptions = new() { MaxDegreeOfParallelism = maxDegreeOfParallelism };
        string message = $") Building LocationContainers Face Files Collection {sourceClass} - {totalSeconds} total second(s)";
        compare.ConstructProgressBar(exifDirectories.Count, message);
        _ = Parallel.For(0, exifDirectories.Count, parallelOptions, (i, state) =>
        {
            compare.Tick();
            LocationContainersParallelFor(distanceSettings, compareSettings, results, exifDirectories[i]);
        });
        return results.AsReadOnly();
    }

}